from arrest import Service
from .resources import {{ resource_imports|sort|join(', ') }}
{% for service in services %}
{{ service.service_id|lower }} = Service(
    name="{{ service.name|lower }}",
    url="{{ service.url }}",
    resources=[{{ service.resources|sort|join(', ') }}]
)
{% endfor %}
